//
// Copyright (c) 2015 Assured Information Security, Inc
//
// Dates Modified:
//  - 4/8/2015: Initial commit
//    Rian Quinn <quinnr@ainfosec.com>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include "qemu/osdep.h"
#include "qemu-common.h"
#include "audio.h"
#include "qemu/timer.h"

#define AUDIO_CAP "openxtaudio"
#include "audio_int.h"

#include "alsadef.h"
#include "openxtargo.h"
#include "openxtdebug.h"
#include "openxtpackets.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// Macros                                                                                              //
/////////////////////////////////////////////////////////////////////////////////////////////////////////

// The following defines the sample buffer that QEMU will use. If there is
// problems with buffering, increasing this value might help, but it might
// also increase latency.
#define MAX_SAMPLES 1024

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// Opaque Structures                                                                                   //
/////////////////////////////////////////////////////////////////////////////////////////////////////////

// Opaque out pointer
typedef struct OpenXTOut {
    HWVoiceOut hw;

} OpenXTOut;

// Opaque in pointer
typedef struct OpenXTIn {
    HWVoiceIn hw;

} OpenXTIn;

bool playback_initialized = false;
bool capture_initialized = false;

// Global Argo Packets
ArgoPacket snd_packet;
ArgoPacket rcv_packet;

// GLobal Argo Connection
ArgoConnection *conn = NULL;

// Global Argo Packet Playback Bodies
OpenXTPlaybackPacket *playback_packet = NULL;
OpenXTPlaybackInitAckPacket *playback_init_ack_packet = NULL;
OpenXTPlaybackSetVolumePacket *playback_set_volume_packet = NULL;
OpenXTPlaybackGetAvailableAckPacket *playback_get_available_ack_packet = NULL;

// Global Argo Packet Capture Bodies
OpenXTCapturePacket *capture_packet = NULL;
OpenXTCaptureAckPacket *capture_ack_packet = NULL;
OpenXTCaptureInitAckPacket *capture_init_ack_packet = NULL;
OpenXTCaptureGetAvailableAckPacket *capture_get_available_ack_packet = NULL;

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// Helpers                                                                                             //
/////////////////////////////////////////////////////////////////////////////////////////////////////////

///
/// Send a blank packet with a specific opcode.
///
/// @desc This function should be used anytime you need to signal to the audio helper,
/// but you don't actually have to send any information minus the opcode itself.
///
/// @param opcode the opcode that you wish to send. This can be any integer, but it should
/// match what you have in the audio helper.
///
/// @return -ENODEV if Argo is not connected
///         negative error code on error
///         0 on success
///
static int openxt_send_blank_packet(int32_t opcode)
{
    int ret;

    // Sanity checks
    openxt_assert_quiet(openxt_argo_isconnected(conn) == true, -ENODEV);

    // Send the blank packet
    ret = openxt_argo_set_opcode(&snd_packet, opcode);
    openxt_assert(ret == 0, ret);
    ret = openxt_argo_set_length(&snd_packet, sizeof(OpenBlankPacket));
    openxt_assert(ret == 0, ret);
    ret = openxt_argo_send(conn, &snd_packet);
    openxt_assert(ret == sizeof(OpenBlankPacket), ret);

    // Success
    return 0;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// Playback                                                                                            //
/////////////////////////////////////////////////////////////////////////////////////////////////////////

///
/// Initializes the playback device. This is one of QEMU's required functions.
///
/// @param hw points to a structure that QEMU provides. You need to initialize
/// this structure by running audio_pcm_init_info, and then setting the # of
/// samples after words.
/// @param as settings structure. Provide information about the audio device
/// as QEMU would like it. When you initialize hw, you can change these settings
/// to whatever you want.
///
/// @return -EINVAL if NULL is passed to the parameters
///         -ENODEV if Argo is not connected or the audio helper fails
///         negative error code on error
///         0 on success
///
static int openxt_playback_init(HWVoiceOut *hw, struct audsettings *as,
                                void *drv_opaque)
{
    int ret;
    struct audsettings settings;

    // Sanity checks
    openxt_checkp(hw, -EINVAL);
    openxt_checkp(as, -EINVAL);
    openxt_assert_quiet(playback_initialized == false, -ENODEV);
    openxt_assert_quiet(openxt_argo_isconnected(conn) == true, -ENODEV);

    // For whatever reason, QEMU will run this function once, or it can run it
    // 1000 times if an error occurs. The result if the VM gets bogged down
    // like mad. This code with the assert above just makes sure that this code
    // can only execute once.... like it should be in the first place.
    playback_initialized = true;

    // Ask the audio helper for the settings that we should be using. Note that
    // we do not get these from QEMU, but instead let the audio helper tell us
    // what it wants to do, and then tell QEMU what that is, which is a little
    // different then what you will see in the alsaaudio driver.
    ret = openxt_send_blank_packet(OPENXT_PLAYBACK_INIT);
    openxt_assert(ret == 0, ret);
    ret = openxt_argo_recv(conn, &rcv_packet);
    openxt_assert(ret == sizeof(OpenXTPlaybackInitAckPacket), ret);

    // Make sure that ALSA configured correctly
    openxt_assert(playback_init_ack_packet->valid != 0, -ENODEV);

    // Set the settings structure that QEMU is expecting with the results
    // from the audio helper. Note that this requires some conversions as
    // ALSA stores the endianness and format together, while QEMU keeps
    // these apart.
    settings.freq = playback_init_ack_packet->freq;
    settings.nchannels = playback_init_ack_packet->nchannels;
    alsa_to_audfmt(playback_init_ack_packet->fmt, &settings.fmt, &settings.endianness);

    // Setup QEMU
    audio_pcm_init_info(&hw->info, &settings);
    hw->samples = 1024;

    // Success
    return 0;
}

///
/// Finalizes the playback device. This is one of QEMU's required functions.
///
/// @param hw points to a structure that QEMU provides.
///
static void openxt_playback_fini(HWVoiceOut *hw)
{
    openxt_send_blank_packet(OPENXT_PLAYBACK_FINI);
}

///
/// Processes the playback device. This is one of QEMU's required functions.
///
/// @param hw points to a structure that QEMU provides.
/// @param live the total number of samples that are available. Attempts to
/// read all of theses samples at once will fail with some strange issues.
/// In general, you need a way to "rate" how many samples you decide to process
/// (i.e. you simply cannot process all of the samples that "live" would suggest
/// can be processed). Spice and the noaudio code do this by using a clock.
/// The ALSA code does this by requesting how many samples the physical hardware
/// can actually process, which seems to work a lot better.
///
/// @return the total number of samples that were processed.
///
static int openxt_playback(HWVoiceOut *hw, int live)
{
    int ret;
    int decr;
    int samples;

    // Sanity checks
    openxt_checkp(hw, 0);
    openxt_assert_quiet(openxt_argo_isconnected(conn) == true, 0);

    // We need to find out how much space is available in ALSA so that we
    // know how many samples to read.
    ret = openxt_send_blank_packet(OPENXT_PLAYBACK_GET_AVAILABLE);
    openxt_assert(ret == 0, 0);
    ret = openxt_argo_recv(conn, &rcv_packet);
    openxt_assert(ret == sizeof(OpenXTPlaybackGetAvailableAckPacket), 0);

    // The following figures out how many samples we "can" process in this
    // iteration (as this function is called by a timer in QEMU). Basically,
    // that total number of samples we can process is the lower value of either
    // the total number of samples provided to the hardware (live), and the
    // total number of samples that ALSA can currently take.
    decr = audio_MIN(live, playback_get_available_ack_packet->available);

    // Now that we know how many samples we can send to ALSA to process, we
    // need to package these samples up into packets and send them to the
    // helper. This loop has to do a couple of things, including making sure
    // that we do not overrun the circular buffer that QEMU maintains, as well
    // as the making sure that we do not overrun the packet itself, as it
    // can only hold a limited number of samples.
    for (samples = decr; samples > 0; )
    {
        int len = samples;

        // Make sure that we are not overrunning the circular buffer that QEMU
        // has, as well as make sure that we are not overrunning the packet
        // buffer.
        len = audio_MIN(len, hw->samples - hw->rpos);
        len = audio_MIN(len, MAX_PCM_BUFFER_SIZE / sizeof(uint32_t));

        // Setup the packet.
        ret = openxt_argo_set_opcode(&snd_packet, OPENXT_PLAYBACK);
        openxt_assert(ret == 0, 0);
        ret = openxt_argo_set_length(&snd_packet, PLAYBACK_PACKET_LENGTH(len));
        openxt_assert(ret == 0, 0);

        // Fill in the packet's body. Note that the hw->clip function is
        // basically a memcpy, the only difference is that len is in samples
        // instead of bytes. The source is our packet, and it's unit is expected
        // to be in 16 bit little endian with 2 channels, which means that
        // each sample is 32 bits (size of a unit32_t). The destination is
        // an array of st_samples which is 64 bits per channel, for a total
        // of 16 bytes per samples. This function does the copy, and conversion
        // for us.
        playback_packet->num_samples = len;
        hw->clip(playback_packet->samples, hw->mix_buf + hw->rpos, len);

        // Send the packet to the helper.
        ret = openxt_argo_send(conn, &snd_packet);
        openxt_assert(ret == PLAYBACK_PACKET_LENGTH(len), 0);

        // We need to advance the internal pointers.
        samples -= len;
        hw->rpos = (hw->rpos + len) % hw->samples;
    }

    // Success
    return decr;
}

///
/// This is one of QEMU's required functions. All of the other drivers
/// just pass this to QEMU's built in function. No idea what it actually
/// does.
///
/// @param sw N/A
/// @param buf N/A
/// @param len N/A
///
/// @return N/A
///
static int openxt_write(SWVoiceOut *sw, void *buf, int len)
{
    return audio_pcm_sw_write(sw, buf, len);
}

///
/// This function returns control information from QEMU.
///
/// @desc This is actually a really important function. Whenever the guest is
/// about to start playing sound, it will first start the audio device, and
/// once it is done, it will "close" the audio device. An easy way to
/// see this is open Windows Media player, and play a song. Once you hit play
/// you will see VOICE_ENABLE come through, and then when the song is done
/// you will see VOICE_DISABLE. ALSA does this with "prepare" and "drop".
/// The other thing that this does is volume control. The value that you get
/// needs to be converted. Basically, QEMU will report a number between
/// 0 and UINT_MAX. That allows you to calculate a linear percentage of what
/// the volume should be. There are some quirks that took a while to work
/// though:
/// - Windows (AC97) never sets the volume to 100% within QEMU. It will
///   always stop at 75%. The running theory is that Windows will not set
///   the gain beyond the hardware's ability to raise the volume without
///   distortion.
/// - Linux alsamixer will not match this percentage. It appears that
///   alsamixer uses the hardware's gain to create a fake logarithmic scale
///   that doesn't match what you see in the hardware itself. Since the gain
///   on the physical hardware will never match what the guest is reporting,
///   alsamixer within the guest and dom0 will not match
/// - Linux amixer will match between the guest and dom0, as well as the
///   percentage that you see QEMU reporting. Amixer doesn't seem to be
///   doing anything strange with the percentage, so everything matches up
///   fine.
/// - Windows (hda) never calls VOICE_VOLUME. It appears to be doing some
///   sort of software volume mixing, which might explain why it doesn't work
///   well, as that operation is pretty slow.
///
/// @param hw points to a structure that QEMU provides.
/// @param cmd that command that QEMU is sending
/// @param ... needs to be parsed based on the command. Look at the other
/// drivers for examples.
///
/// @return -1 on not supported or error
///         0 success / supported
///
static int openxt_playback_ctl(HWVoiceOut *hw, int cmd, ...)
{
    int ret;

    // Sanity checks
    openxt_assert_quiet(openxt_argo_isconnected(conn) == true, -ENODEV);

    // Process the command
    switch(cmd) {

        case VOICE_ENABLE:
            ret = openxt_send_blank_packet(OPENXT_PLAYBACK_ENABLE_VOICE);
            openxt_assert(ret == 0, -1);

            // Success
            return 0;

        case VOICE_DISABLE:
            ret = openxt_send_blank_packet(OPENXT_PLAYBACK_DISABLE_VOICE);
            openxt_assert(ret == 0, -1);

            // Success
            return 0;

        case VOICE_VOLUME:
        {
            va_list va;
            SWVoiceOut *sw = NULL;

            // Get the SWVoiceOut from the args
            va_start(va, cmd);
            sw = va_arg(va, SWVoiceOut*);
            va_end(va);

            // Make sure that we actually got an argument back.
            if (sw != NULL) {

                // Setup the packet
                ret = openxt_argo_set_opcode(&snd_packet, OPENXT_PLAYBACK_SET_VOLUME);
                openxt_assert(ret == 0, -1);
                ret = openxt_argo_set_length(&snd_packet, sizeof(OpenXTPlaybackSetVolumePacket));
                openxt_assert(ret == 0, -1);

                // Setup the packet.
                if (sw->vol.mute) {

                    playback_set_volume_packet->vol = 0;
                    playback_set_volume_packet->enabled = 0;

                } else {

                    playback_set_volume_packet->vol = (sw->vol.l * 100) / UINT_MAX;
                    playback_set_volume_packet->enabled = 1;
                }

                // Send the packet
                ret = openxt_argo_send(conn, &snd_packet);
                openxt_assert(ret == sizeof(OpenXTPlaybackSetVolumePacket), -1);

                // Done
                return 0;
            }

            // Failure
            return -1;
        }

        default:
            return -1;
    }
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// Capture                                                                                             //
/////////////////////////////////////////////////////////////////////////////////////////////////////////

///
/// Initializes the capture device. This is one of QEMU's required functions.
///
/// @param hw points to a structure that QEMU provides. You need to initialize
/// this structure by running audio_pcm_init_info, and then setting the # of
/// samples after words.
/// @param as settings structure. Provide information about the audio device
/// as QEMU would like it. When you initialize hw, you can change these settings
/// to whatever you want.
///
/// @return -EINVAL if NULL is passed to the parameters
///         -ENODEV if Argo is not connected or the audio helper fails
///         negative error code on error
///         0 on success
///
static int openxt_capture_init(HWVoiceIn *hw, struct audsettings *as,
                               void *drv_opaque)
{
    int ret;
    struct audsettings settings;

    // Sanity checks
    openxt_checkp(hw, -EINVAL);
    openxt_checkp(as, -EINVAL);
    openxt_assert_quiet(capture_initialized == false, -ENODEV);
    openxt_assert_quiet(openxt_argo_isconnected(conn) == true, -ENODEV);

    // For whatever reason, QEMU will run this function once, or it can run it
    // 1000 times if an error occurs. The result if the VM gets bogged down
    // like mad. This code with the assert above just makes sure that this code
    // can only execute once.... like it should be in the first place.
    capture_initialized = true;

    // Ask the audio helper for the settings that we should be using. Note that
    // we do not get these from QEMU, but instead let the audio helper tell us
    // what it wants to do, and then tell QEMU what that is, which is a little
    // different then what you will see in the alsaaudio driver.
    ret = openxt_send_blank_packet(OPENXT_CAPTURE_INIT);
    openxt_assert(ret == 0, ret);
    ret = openxt_argo_recv(conn, &rcv_packet);
    openxt_assert(ret == sizeof(OpenXTCaptureInitAckPacket), ret);

    // Make sure that ALSA configured correctly
    openxt_assert(capture_init_ack_packet->valid != 0, -ENODEV);

    // Set the settings structure that QEMU is expecting with the results
    // from the audio helper. Note that this requires some conversions as
    // ALSA stores the endianness and format together, while QEMU keeps
    // these apart.
    settings.freq = capture_init_ack_packet->freq;
    settings.nchannels = capture_init_ack_packet->nchannels;
    alsa_to_audfmt(capture_init_ack_packet->fmt, &settings.fmt, &settings.endianness);

    // Setup QEMU
    audio_pcm_init_info(&hw->info, &settings);
    hw->samples = 1024;

    // Success
    return 0;
}

///
/// Finalizes the capture device. This is one of QEMU's required functions.
///
/// @param hw points to a structure that QEMU provides.
///
static void openxt_capture_fini(HWVoiceIn *hw)
{
    openxt_send_blank_packet(OPENXT_CAPTURE_FINI);
}

///
/// Processes the capture device. This is one of QEMU's required functions.
///
/// @param hw points to a structure that QEMU provides.
/// @return the total number of samples that were processed.
///
static int openxt_capture(HWVoiceIn *hw)
{
    int ret;
    int decr;
    int samples;
    int written = 0;

    // Sanity checks
    openxt_checkp(hw, 0);
    openxt_assert_quiet(openxt_argo_isconnected(conn) == true, 0);

    // Calculate the total amount of space that is available in the emulated
    // hardware. This is basically the total size of the buffer, minus the
    // number of samples that are "live" in the buffer. Note that we use the
    // max function as it is possible to get a negative here, so we check for
    // that.
    decr = audio_MAX(hw->samples - audio_pcm_hw_get_live_in(hw), 0);

    // Loop through all of the samples that need to be processed, get them
    // from don0, and provide them to the emulated hardware.
    for (samples = decr; samples > 0; )
    {
        int len = samples;

        // Make sure that we are not overrunning the circular buffer that QEMU
        // has, as well as make sure that we are not overrunning the packet
        // buffer.
        len = audio_MIN(len, hw->samples - hw->wpos);
        len = audio_MIN(len, MAX_PCM_BUFFER_SIZE / sizeof(uint32_t));

        // Setup the packet so that we can ask for a set of samples from dom0
        ret = openxt_argo_set_opcode(&snd_packet, OPENXT_CAPTURE);
        openxt_assert(ret == 0, 0);
        ret = openxt_argo_set_length(&snd_packet, sizeof(OpenXTCapturePacket));
        openxt_assert(ret == 0, 0);

        // Tell dom0 how many samples need to be collected.
        capture_packet->num_samples = len;

        // Send the request to dom0, and then wait for the samples to come back.
        // Note that in this case, we don't really have a way to validate the
        // packet size here for the result. The Argo logic will at least verify
        // that the packet's len is what it reports, but it's possible for a
        // messed up audio helper to give use a num_samples that doesn't correlate
        // with the actually number of samples. This will not result in an overflow
        // but it would result in garbage being sent to QEMU (not a big deal).
        ret = openxt_argo_send(conn, &snd_packet);
        openxt_assert(ret == sizeof(OpenXTCapturePacket), 0);
        ret = openxt_argo_recv(conn, &rcv_packet);
        openxt_assert(ret >= 0, 0);

        // Check to see if we got any samples back. If we didn't there is nothing
        // to do, and we can stop
        if (capture_ack_packet->num_samples == 0)
            break;

        // Fill in the packet's body. Note that the hw->conv function is
        // basically a memcpy, the only difference is that len is in samples
        // instead of bytes. The destination is the current position in the circular
        // buffer that QEMU provides, and the source is our packet.
        hw->conv(hw->conv_buf + hw->wpos, capture_ack_packet->samples, capture_ack_packet->num_samples);

        // We need to advance the internal pointers.
        samples -= capture_ack_packet->num_samples;
        written += capture_ack_packet->num_samples;
        hw->wpos = (hw->wpos + capture_ack_packet->num_samples) % hw->samples;

        // If the number of samples that we got back does not match the number
        // of samples that we asked for, it means that the source is out of
        // samples, in which case we can stop processing in this iteration.
        if (capture_ack_packet->num_samples != len)
            break;
    }

    // Success
    return written;
}

///
/// This is one of QEMU's required functions. All of the other drivers
/// just pass this to QEMU's built in function. No idea what it actually
/// does.
///
/// @param sw N/A
/// @param buf N/A
/// @param len N/A
///
/// @return N/A
///
static int openxt_read(SWVoiceIn *sw, void *buf, int size)
{
    return audio_pcm_sw_read(sw, buf, size);
}

///
/// This function returns control information from QEMU.
///
/// @desc This is actually a really important function. Whenever the guest is
/// about to start recording sound, it will first start the audio device, and
/// once it is done, it will "close" the audio device. An easy way to
/// see this is open Windows Recorder. As soon as you see recorder start,
/// you will get an VOICE_ENABLE and you will get a VOICE_DISABLE when
/// you close the app. This is important because these controls will prevent
/// the hardware and software from processing when it doesn't need to be.
/// One issue you might notice is ALSA by default in Linux seems to enable
/// the device on bootup, so basically, in Linux, the capture device is
/// firing constantly, even if your not using it. Either disable capture
/// support in the VM's settings, or turn off the capture device in ALSA
/// mixer once you start up the guest.
///
/// @param hw points to a structure that QEMU provides.
/// @param cmd that command that QEMU is sending
/// @param ... needs to be parsed based on the command. Look at the other
/// drivers for examples.
///
/// @return -1 on not supported or error
///         0 success / supported
static int openxt_capture_ctl(HWVoiceIn *hw, int cmd, ...)
{
    int ret;

    // Sanity checks
    openxt_assert_quiet(openxt_argo_isconnected(conn) == true, -ENODEV);

    // Process the command
    switch(cmd) {

        case VOICE_ENABLE:
            ret = openxt_send_blank_packet(OPENXT_CAPTURE_ENABLE_VOICE);
            openxt_assert(ret == 0, -1);

            // Success
            return 0;

        case VOICE_DISABLE:
            ret = openxt_send_blank_packet(OPENXT_CAPTURE_DISABLE_VOICE);
            openxt_assert(ret == 0, -1);

            // Success
            return 0;

        case VOICE_VOLUME:

            // TODO: We need to implement capture audio volume support. To support this
            // a lot of work needs to be done to the audio helper side, as only one VM
            // should have access to the capture device at a time, which it doesn't do
            // now.
            return -1;

        default:
            return -1;
    }
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// Init / Fini                                                                                         //
/////////////////////////////////////////////////////////////////////////////////////////////////////////

///
/// Initializes the audio front driver
///
/// @return an opaque pointer. We don't use this.
///
static void *openxt_audio_init(void)
{
    // Setup
    openxt_debug_init();
    conn = openxt_argo_open(XEN_ARGO_PORT_NONE, XEN_ARGO_DOMID_ANY, OPENXT_AUDIO_PORT, 0);
    openxt_assert(conn != NULL, NULL);

    // Cleanup memory(safety)
    memset(&snd_packet, 0, sizeof(ArgoPacket));
    memset(&rcv_packet, 0, sizeof(ArgoPacket));

    // Pointer checks
    openxt_checkp(playback_packet = openxt_argo_get_body(&snd_packet), NULL);
    openxt_checkp(playback_init_ack_packet = openxt_argo_get_body(&rcv_packet), NULL);
    openxt_checkp(playback_set_volume_packet = openxt_argo_get_body(&snd_packet), NULL);
    openxt_checkp(playback_get_available_ack_packet = openxt_argo_get_body(&rcv_packet), NULL);

    // Pointer checks
    openxt_checkp(capture_packet = openxt_argo_get_body(&snd_packet), NULL);
    openxt_checkp(capture_ack_packet = openxt_argo_get_body(&rcv_packet), NULL);
    openxt_checkp(capture_init_ack_packet = openxt_argo_get_body(&rcv_packet), NULL);
    openxt_checkp(capture_get_available_ack_packet = openxt_argo_get_body(&rcv_packet), NULL);

    // Size checks
    openxt_assert(openxt_argo_validate(sizeof(OpenXTPlaybackPacket)) == true, NULL);
    openxt_assert(openxt_argo_validate(sizeof(OpenXTPlaybackInitAckPacket)) == true, NULL);
    openxt_assert(openxt_argo_validate(sizeof(OpenXTPlaybackSetVolumePacket)) == true, NULL);
    openxt_assert(openxt_argo_validate(sizeof(OpenXTPlaybackGetAvailableAckPacket)) == true, NULL);

    // Size checks
    openxt_assert(openxt_argo_validate(sizeof(OpenXTCapturePacket)) == true, NULL);
    openxt_assert(openxt_argo_validate(sizeof(OpenXTCaptureAckPacket)) == true, NULL);
    openxt_assert(openxt_argo_validate(sizeof(OpenXTCaptureInitAckPacket)) == true, NULL);
    openxt_assert(openxt_argo_validate(sizeof(OpenXTCaptureGetAvailableAckPacket)) == true, NULL);

    // Success. We don't have an init/fini pointer to return as we use
    // global memory like the other drivers, so like the other drivers we
    // just pass the address of this function, as returning NULL indicates
    // and error to QEMU.
    return &openxt_audio_init;
}

///
/// Finalize the audio front driver
///
static void openxt_audio_fini(void *opaque)
{
    int ret;

    // Unused / Sanity checks
    openxt_unused(opaque);
    openxt_assert_quiet(openxt_argo_isconnected(conn) == true);

    // Setup the packet
    ret = openxt_send_blank_packet(OPENXT_FINI);
    openxt_assert(ret == 0);

    // Cleanup
    openxt_argo_close(conn);
    openxt_debug_fini();
}

static struct audio_pcm_ops openxt_pcm_ops = {
    .init_out = openxt_playback_init,
    .fini_out = openxt_playback_fini,
    .run_out  = openxt_playback,
    .write    = openxt_write,
    .ctl_out  = openxt_playback_ctl,

    .init_in  = openxt_capture_init,
    .fini_in  = openxt_capture_fini,
    .run_in   = openxt_capture,
    .read     = openxt_read,
    .ctl_in   = openxt_capture_ctl
};

static struct audio_driver openxt_audio_driver = {
    .name           = "openxt",
    .descr          = "OpenXT Audio Driver",
    .options        = NULL,
    .init           = openxt_audio_init,
    .fini           = openxt_audio_fini,
    .pcm_ops        = &openxt_pcm_ops,
    .can_be_default = 1,
    .max_voices_out = INT_MAX,
    .max_voices_in  = INT_MAX,
    .voice_size_out = sizeof(OpenXTOut),
    .voice_size_in  = sizeof(OpenXTIn)
};

static void register_audio_openxt(void)
{
    audio_driver_register(&openxt_audio_driver);
}
type_init(register_audio_openxt);
